from collections.abc import Callable
from typing import TYPE_CHECKING

from transitions import Machine

from django.db import models
from django.utils import timezone

from django_model_fsm.helpers import abstract_method
from django_model_fsm.dataclasses import StateData, WorkflowData
from django_model_fsm.fields import TransitionField

if TYPE_CHECKING:
    from accounts.models import User


class SupportObj:
    pass


class TransitionsMixin(models.Model):
    workflow_data: WorkflowData = None
    workflow_callbacks: list[str] = []

    transition_state = models.TransitionField(
        "Transition's state stored in database",
        blank=True,
        null=False,
        max_length=100,
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        by: User = kwargs.pop("saved_by", None)

        self.transition_state = self.transition_support.state

        if "history" not in self.transition_history:
            self.transition_history["history"] = []

        if self.pk:
            try:
                label_from = self.get_state_as_label(
                    state=self.transition_state_initial_value,
                )
            except Exception:  # NOQA: BLE001
                label_from = ""

            try:
                label_to = self.get_state_as_label(state=self.transition_state)
            except Exception:  # NOQA: BLE001
                label_to = ""

            if self.transition_state_initial_value != self.transition_state:
                self.transition_history["history"].append(
                    {
                        "by": getattr(by, "email", None) if by is not None else None,
                        "date": int(timezone.now().timestamp() * 1000),
                        "state_from": self.transition_state_initial_value,
                        "state_to": self.transition_state,
                        "label_from": label_from,
                        "label_to": label_to,
                    },
                )

        return super().save(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.transition_state:
            self.transition_state = self.workflow_data.initial_state.id

        self.transition_state_initial_value = self.transition_state

        self.__init_support_object_with_callbacks()

        self.fsm = Machine(
            self.transition_support,
            states=self.workflow_data.states,
            transitions=self.workflow_data.transitions,
            initial=self.transition_state,
        )

    def __init_support_object_with_callbacks(self)-> None:
        self.transition_support = SupportObj()
        self.__workflow_add_callbacks()

    def __workflow_add_callbacks(self) -> None:
        for callback in self.workflow_callbacks:
            setattr(self.transition_support, callback, getattr(self, callback))

    def workflow_get_accessible_triggers(self) -> list[str]:
        return self.workflow_data.get_triggers_from_state(state=self.workflow_state)

    def workflow_get_accessible_triggers_with_labels(self) -> list[dict]:
        return self.workflow_data.get_triggers_and_labels_from_state(
            state=self.workflow_state,
        )

    @classmethod
    @property
    def workflow_all_states_data(cls):
        return cls.workflow_data.states_data

    @property
    def workflow_state(self):
        return self.transition_support.state

    @property
    def workflow_state_as_label(self) -> str:
        return self.workflow_data.get_state_as_label(state=self.workflow_state)

    def is_workflow_state_equal(self, state: StateData) -> bool:
        return self.workflow_state == state.id

    @property
    def workflow_trigger(self) -> Callable:
        return self.transition_support.trigger



from dataclasses import dataclass


@dataclass
class TriggerData:
    id: str  # NOQA: A003
    label: str
    order: int = 0
    need_confirmation: bool = False

    def __str__(self):
        return self.id
    



from django.db import models


class TransitionField(models.CharField):
    description = "Transition state"

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = kwargs["max_length"] if "max_length" in kwargs else 255
        kwargs["blank"] = True
        kwargs["null"] = False

        super().__init__(*args, **kwargs)



from dataclasses import dataclass

from transitions import State

from .state_data import StateData
from .trigger_data import TriggerData


@dataclass
class WorkflowData:
    initial_state: str
    transitions_data: list[dict]

    @property
    def labels(self) -> dict:
        return {x.id: x.label for x in self.states_data}

    @property
    def states(self) -> list[State]:
        return [State(name=x.id) for x in self.states_data]

    @property
    def states_data(self) -> list[StateData]:
        states = []
        for x in self.transitions_data:
            if x["source"] not in states:
                states.append(x["source"])
            if x["dest"] not in states:
                states.append(x["dest"])
        return states

    @property
    def triggers(self) -> dict:
        return {x.id: x for x in self.triggers_data}

    @property
    def triggers_data(self) -> list[TriggerData]:
        triggers = []
        for x in self.transitions_data:
            if x["trigger"] not in triggers:
                triggers.append(x["trigger"])

        return triggers

    def get_labels_for_triggers(self) -> dict:
        return {key: value.label for key, value in self.triggers.items()}

    def get_state_as_label(self, state: str) -> str:
        return self.labels[state]

    def get_triggers_and_labels_from_state(self, state: str) -> list[dict]:
        ordered_triggers = self.get_triggers_from_state(state=state)
        return {
            x: {
                "label": self.triggers[x].label,
                "need_confirmation": self.triggers[x].need_confirmation,
            }
            for x in ordered_triggers
        }

    def get_triggers_from_state(self, state: str) -> list[str]:
        triggers = [x["trigger"] for x in self.transitions if x["source"] == state]
        # ordered triggers
        return [
            x[0]
            for x in sorted(
                [(x, self.triggers[x].order) for x in triggers],
                key=lambda x: x[1],
            )
        ]

    @property
    def transitions(self) -> list[dict]:
        return [
            {key: x[key].id if key in ["trigger", "source", "dest"] else x[key] for key in x}
            for x in self.transitions_data
        ]
